=== modified file 'Makefile'
--- Makefile	2012-01-11 08:06:03 +0000
+++ Makefile	2012-04-20 02:20:58 +0000
@@ -1,4 +1,4 @@
 check:
-	python -m unittest grackle.tests.test_client
+	python -m testtools.run grackle.tests.test_suite
 
 .PHONY: check

=== added file 'grackle-create-instance'
--- grackle-create-instance	1970-01-01 00:00:00 +0000
+++ grackle-create-instance	2012-01-22 10:39:15 +0000
@@ -0,0 +1,23 @@
+#!/usr/bin/env python
+
+from optparse import OptionParser
+
+from grackle.model import create_schema
+
+
+parser = OptionParser(usage="usage: %prog [options] <hostname> <keyspace>")
+parser.add_option(
+    "--clobber", action="store_true", default=False, dest="clobber",
+    help="clear the keyspace before initialising it")
+parser.add_option(
+    "--create-keyspace", action="store_true", default=False,
+    dest="create_keyspace",
+    help="create the keyspace (with SimpleStrategy, RF=1)")
+options, args = parser.parse_args()
+
+if len(args) != 2:
+    parser.error("hostname and keyspace required")
+
+create_schema(
+    args[0], args[1], clobber=options.clobber,
+    create_keyspace=options.create_keyspace)

=== modified file 'grackle-put-message'
--- grackle-put-message	2012-01-13 13:52:09 +0000
+++ grackle-put-message	2012-01-30 18:30:54 +0000
@@ -2,9 +2,11 @@
 import sys
 from grackle import client
 
+
 class UserError(Exception):
     """An error caused by a user mistake."""
 
+
 def main(*args):
     if len(args) != 2:
         raise UserError("Wrong number of arguments supplied.")

=== added file 'grackle/cassandra.py'
--- grackle/cassandra.py	1970-01-01 00:00:00 +0000
+++ grackle/cassandra.py	2012-01-22 10:39:15 +0000
@@ -0,0 +1,41 @@
+# Copyright (c) 2011 Canonical Ltd
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Affero General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Affero General Public License for more details.
+#
+# You should have received a copy of the GNU Affero General Public
+# License along with this program. If not, see
+# <http://www.gnu.org/licenses/>.
+
+"""Things to ease working with cassandra."""
+
+from pycassa.pool import InvalidRequestError
+
+
+def workaround_1779(callable, *args, **kwargs):
+    """Workaround cassandra not being able to do concurrent schema edits.
+
+    The callable is tried until it does not raised InvalidRequestException
+    with why = "Previous version mismatch. cannot apply."
+
+    :param callable: The callable to call.
+    :param args: The args for it.
+    :param kwargs: The kwargs for it.
+    :return: The result of calling the callable.
+    """
+    while True:
+        # Workaround https://issues.apache.org/jira/browse/CASSANDRA-1779:
+        # Cassandra cannot do concurrent schema changes.
+        try:
+            return callable(*args, **kwargs)
+            break
+        except InvalidRequestError as e:
+            if e.why != 'Previous version mismatch. cannot apply.':
+                raise

=== modified file 'grackle/client.py'
--- grackle/client.py	2012-01-30 18:28:17 +0000
+++ grackle/client.py	2012-04-20 02:20:58 +0000
@@ -1,3 +1,5 @@
+__metaclass__ = type
+
 import httplib
 import simplejson
 from urlparse import urlunparse
@@ -6,9 +8,13 @@
     urlencode,
 )
 
-
-class UnsupportedOrder(Exception):
-    """Raised when an Unsupported order is requested."""
+from grackle.error import (
+    ArchiveIdExists,
+    MessageIdNotFound,
+    UnparsableDateRange,
+    UnsupportedDisplayType,
+    UnsupportedOrder,
+    )
 
 
 class GrackleClient:
@@ -24,26 +30,45 @@
         self.port = port
         self.netloc = '%s:%d' % (host, port)
 
-    def archive_url(self, archive_id, query):
+    def archive_url(self, path, query):
         """Return the URL for an archive
 
-        :param archive_id: The id of the archive to generate the URL for.
+        :param path: The path to generate the URL for.
+            Maybe be '', 'archive_id', or 'archive_id/message_id'
         :param query: The query to use in the URL, as a dict.
         """
-        path = '/archive/%s' % quote(archive_id)
+        path = '/archive/%s' % quote(path)
         query_string = urlencode(query)
         return urlunparse(('http', self.netloc, path, '', query_string, ''))
 
     def _get_connection(self):
         return httplib.HTTPConnection(self.host, self.port)
 
-    def _method_archive(self, method, archive_id, query, body=None):
+    def _method_archive(self, method, path, query, body=None):
         """Perform an HTTP method on an archive's URL."""
-        url = self.archive_url(archive_id, query)
+        url = self.archive_url(path, query)
         connection = self._get_connection()
         connection.request(method, url, body)
         return connection.getresponse()
 
+    def put_archive(self, archive_id, mbox=None):
+        """Create an archive.
+
+        :param archive_id: The archive id.
+        :param mbox: An optional mbox with messages to add to the new archive.
+        """
+        response = self._method_archive(
+            'PUT', archive_id, {}, None)
+        response.read()
+        if response.status == httplib.BAD_REQUEST:
+            if response.reason == ArchiveIdExists.__doc__:
+                raise ArchiveIdExists
+            raise Exception('wtf')
+        elif response.status == httplib.CREATED:
+            return
+        else:
+            raise Exception('!!')
+
     def put_message(self, archive_id, key, file_obj):
         """Put a message into an archive.
 
@@ -52,23 +77,31 @@
             the message.
         :param file_obj: The raw text of the message, as a file.
         """
+        path = '%s/%s' % (archive_id, key)
         response = self._method_archive(
-            'POST', archive_id, {'key': key}, file_obj.read())
+            'PUT', path, {}, file_obj.read())
         response.read()
         if response.status == httplib.BAD_REQUEST:
+            if response.reason == ArchiveIdExists.__doc__:
+                raise ArchiveIdExists
             raise Exception('wtf')
         elif response.status == httplib.CREATED:
             return
         else:
             raise Exception('!!')
 
-    def get_messages(self, archive_id, message_ids=None, limit=None,
-                     memo=None, order=None, headers=None,
-                     max_body_length=None, include_hidden=False):
+    def get_messages(self, archive_id, message_ids=None, date_range=None,
+                     limit=None, memo=None, order=None, headers=None,
+                     include_hidden=False, max_body_length=None,
+                     display_type='all'):
         """Retrieve specified messages.
 
         :param archive_id: The archive to retrieve messages from.
         :param message_ids: (optional) Retrieve only messages with these ids.
+        :param date_range: Retrieve the messages from or between a range of
+            dates. Example: 2012-01-01..2012-01-31 retrieve all the messages
+            between the 01 and 31 of January, including message from 01
+            and 31.
         :param limit: The maximum number of messages to return.  The server
             may, at its discretion, return fewer.
         :param memo: (optional) Opaque identifier describing the position in
@@ -87,10 +120,17 @@
             bodies.
         :param include_hidden: If true, include messages that have been
             flagged "hidden" in the results.
+        :param display_type: Adjust the message content to meet the needs of
+            the intended display. Valid values are:
+            all: (the default) include all message content.
+            text-only: include only plain/text parts; exclude all other parts.
+            headers-only: include only the message headers.
         """
         parameters = {}
         if message_ids is not None:
             parameters['message_ids'] = message_ids
+        if date_range is not None:
+            parameters['date_range'] = date_range
         if limit is not None:
             parameters['limit'] = limit
         if memo is not None:
@@ -101,10 +141,31 @@
             parameters['headers'] = headers
         if max_body_length is not None:
             parameters['max_body_length'] = max_body_length
+        parameters['display_type'] = display_type
         parameters['include_hidden'] = include_hidden
         query = {'parameters': simplejson.dumps(parameters)}
         response = self._method_archive('GET', archive_id, query)
         if response.status == httplib.BAD_REQUEST:
-            raise UnsupportedOrder
+            if response.reason == UnsupportedOrder.__doc__:
+                raise UnsupportedOrder
+            elif response.reason == UnsupportedDisplayType.__doc__:
+                raise UnsupportedDisplayType
+            elif response.reason == UnparsableDateRange.__doc__:
+                raise UnparsableDateRange
+            else:
+                raise ValueError('Bad request')
+        data = response.read()
+        return simplejson.loads(data)
+
+    def hide_message(self, archive_id, message_id, hidden):
+        parameters = {
+            'hidden': hidden,
+            }
+        query = {'parameters': simplejson.dumps(parameters)}
+        path = '%s/%s' % (archive_id, message_id)
+        response = self._method_archive('POST', path, query)
+        if response.status == httplib.BAD_REQUEST:
+            if response.reason == MessageIdNotFound.__doc__:
+                raise MessageIdNotFound
         data = response.read()
         return simplejson.loads(data)

=== added file 'grackle/error.py'
--- grackle/error.py	1970-01-01 00:00:00 +0000
+++ grackle/error.py	2012-03-17 21:36:11 +0000
@@ -0,0 +1,33 @@
+__metaclass__ = type
+__all__ = [
+    'ArchiveIdExists',
+    'ArchiveIdNotFound',
+    'MessageIdNotFound',
+    'UnparsableDateRange',
+    'UnsupportedDisplayType',
+    'UnsupportedOrder',
+    ]
+
+
+class ArchiveIdExists(Exception):
+    """An archive matching the archive_id was found in the store."""
+
+
+class ArchiveIdNotFound(Exception):
+    """No archive matching the archive_id was found in the store."""
+
+
+class MessageIdNotFound(Exception):
+    """No message matching the message_id was found in the archive."""
+
+
+class UnparsableDateRange(Exception):
+    """The date_range was not in the format of 2012-01-01..2012-01-31."""
+
+
+class UnsupportedDisplayType(Exception):
+    """Raised when an Unsupported display_type is requested."""
+
+
+class UnsupportedOrder(Exception):
+    """Raised when an Unsupported order is requested."""

=== added file 'grackle/model.py'
--- grackle/model.py	1970-01-01 00:00:00 +0000
+++ grackle/model.py	2012-01-24 22:53:32 +0000
@@ -0,0 +1,231 @@
+# Copyright (c) 2012 Canonical Ltd
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Affero General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Affero General Public License for more details.
+#
+# You should have received a copy of the GNU Affero General Public
+# License along with this program. If not, see
+# <http://www.gnu.org/licenses/>.
+
+import calendar
+import datetime
+import email.parser
+from email.utils import parsedate_tz
+import functools
+import logging
+import uuid
+
+import dateutil.tz
+import pycassa.pool
+from pycassa.system_manager import (
+    LEXICAL_UUID_TYPE,
+    SystemManager,
+    TIME_UUID_TYPE,
+    )
+from pycassa.util import convert_time_to_uuid
+
+from grackle.cassandra import workaround_1779
+
+
+def create_schema(host, keyspace, clobber=False, create_keyspace=False):
+    mgr = SystemManager(host)
+
+    if create_keyspace:
+        mgr.create_keyspace(keyspace, replication_factor=1)
+
+    if clobber:
+        for cf in mgr.get_keyspace_column_families(keyspace):
+            mgr.drop_column_family(keyspace, cf)
+
+    try:
+        workaround_1779(
+            mgr.create_column_family, keyspace, 'message',
+            key_validation_class=LEXICAL_UUID_TYPE)
+        workaround_1779(
+            mgr.create_column_family, keyspace, 'archive_message',
+            comparator_type=TIME_UUID_TYPE,
+            default_validation_class=LEXICAL_UUID_TYPE)
+        pass
+    finally:
+        mgr.close()
+
+
+def _parse_message(message):
+    """Get a date and dict of an RFC822 message."""
+    parsed = email.parser.Parser().parsestr(message)
+    message_dict = {}
+
+    for key in ('from', 'to', 'subject', 'message-id'):
+        value = parsed.get(key, None)
+        if value is not None:
+            message_dict[key] = value
+
+    date = parsed.get('date')
+    if date is not None:
+        try:
+            pdate = parsedate_tz(date)
+            date = datetime.datetime(
+                *pdate[:6],
+                tzinfo=dateutil.tz.tzoffset('', pdate[9]))
+        except ValueError:
+            pass
+    message_dict['date'] = date.isoformat() if date is not None else None
+
+    return date, message_dict
+
+
+def _utc_datetime(dt):
+    return dt.astimezone(dateutil.tz.tzutc())
+
+
+def _utc_timestamp(dt):
+    return calendar.timegm(_utc_datetime(dt).timetuple())
+
+
+def _utc_timeuuid(dt, lowest_val=True):
+    return convert_time_to_uuid(_utc_timestamp(dt), lowest_val)
+
+
+def _cmp_timeuuid(a, b):
+    if a.time != b.time:
+        return cmp(a.time, b.time)
+    return cmp(a, b)
+
+
+def _bound_timeuuid(a, b, max=False):
+    if a == '' or _cmp_timeuuid(b, a) == (1 if max else -1):
+        return b
+    return a
+
+
+def _make_bounds(memo, range_start, range_finish, backward):
+    start = finish = ''
+    if memo != '':
+        memo = uuid.UUID(memo)
+    if backward:
+        finish = memo
+    else:
+        start = memo
+    if range_start is not None:
+        start = _bound_timeuuid(
+            start, _utc_timeuuid(range_start), max=True)
+    if range_finish is not None:
+        finish = _bound_timeuuid(
+            finish, _utc_timeuuid(range_finish, lowest_val=False))
+    return memo, start, finish
+
+
+LEGAL_HEADERS = set([
+    'date', 'from', 'subject', 'message-id',
+    ])
+
+
+def _format_message(message, headers=[], include_raw=False):
+    data = {}
+
+    if headers:
+        assert not set(headers).difference(LEGAL_HEADERS)
+        hdict = {}
+        for header in headers:
+            hdict[header] = message.get(header)
+        data['headers'] = hdict
+
+    if include_raw:
+        data['raw'] = message['raw']
+
+    return data
+
+
+class CassandraConnection(object):
+
+    def __init__(self, keyspace, hosts):
+        self._keyspace = keyspace
+        self._hosts = hosts
+        self._connection = self._connect()
+        self._pool = self._connect()
+        self.messages = self._column_family('message')
+        self.archive_messages = self._column_family('archive_message')
+
+    def _connect(self):
+        return pycassa.pool.ConnectionPool(self._keyspace, self._hosts)
+
+    def _column_family(self, name):
+        return pycassa.ColumnFamily(self._pool, name)
+
+    def add_message(self, archive_uuid, message):
+        message_uuid = uuid.uuid4()
+        message_date, message_dict = _parse_message(message)
+        message_dict['raw'] = message
+        message_dict['date_created'] = (
+            datetime.datetime.utcnow().isoformat() + 'Z')
+        self.messages.insert(message_uuid, message_dict)
+        self.archive_messages.insert(
+            archive_uuid,
+            {_utc_timestamp(message_date): message_uuid})
+        logging.debug(
+            'Imported %s into %s'
+            % (message_dict.get('message-id', None), archive_uuid))
+        return message_uuid
+
+    def _trim(self, sequence, end):
+        """Return the sequence with one of the ends trimmed.
+
+        :param end: if true, remove the last element. otherwise remove
+            the first.
+        """
+        if end:
+            return sequence[:-1]
+        else:
+            return sequence[1:]
+
+    def get_messages(self, archive_uuid, order, count, memo, backward=False,
+                     start_date=None, finish_date=None, format='all',
+                     headers=['from', 'date', 'subject', 'message-id']):
+        if order in ("date", "-date"):
+            reversed = order[0] == '-'
+        else:
+            raise AssertionError("Unsupported order.")
+
+        memo, start, finish = _make_bounds(
+            memo, start_date, finish_date, backward)
+
+        # Get up to n+1 messages from the memo: the last item of the
+        # previous batch (because that's where the memo starts) + this
+        # batch.
+        pairs = self.archive_messages.get(
+            archive_uuid, column_count=count + 1, column_start=start,
+            column_finish=finish, column_reversed=reversed).items()
+
+        if len(pairs) and memo and pairs[0][0] <= memo:
+            # The memo (from the previous batch) was included in the result.
+            # Trim it.
+            pairs = self._trim(pairs, False ^ backward)
+        elif len(pairs) > count:
+            # There was no memo in the result, so the n+1th element is
+            # unnecessary. Kill it.
+            pairs = self._trim(pairs, True ^ backward)
+
+        if len(pairs) == 0:
+            return (None, [], None)
+
+        assert 0 < len(pairs) <= count
+
+        # We've narrowed down the message references. Fetch the messages.
+        ids = [v for k, v in pairs]
+        formatter = functools.partial(
+            _format_message, headers=headers, include_raw=True)
+        # XXX: No need to get all columns. Restrict based on format.
+        messages = self.messages.multiget(ids)
+
+        return (
+            str(pairs[0][0]),
+            [formatter(messages[id]) for id in ids],
+            str(pairs[-1][0]),
+            )

=== added file 'grackle/service.py'
--- grackle/service.py	1970-01-01 00:00:00 +0000
+++ grackle/service.py	2012-03-17 23:01:16 +0000
@@ -0,0 +1,189 @@
+__metaclass__ = type
+__all__ = [
+    'ForkedFakeService',
+    'GrackleService',
+    ]
+
+import httplib
+import logging
+import os
+from signal import SIGKILL
+import simplejson
+import sys
+from wsgiref.headers import Headers
+from wsgiref.simple_server import make_server
+from wsgiref.util import shift_path_info
+
+from grackle.store import (
+    MemoryStore,
+    )
+
+
+class GrackleService:
+    """A request handler that forwards to an archive store."""
+
+    def __init__(self, store):
+        self.store = store
+        self.logger = logging.getLogger('http')
+
+    def __call__(self, environ, start_response):
+        self.environ = environ
+        self.start_response = start_response
+        self.headers = Headers([('content-type', 'application/json')])
+        self.method = environ['REQUEST_METHOD']
+        if '://' in environ['PATH_INFO']:
+            # All the needed information is embedded in PATH_INFO.
+            shift_path_info(environ)  # shift the host and or port.
+            self.application = shift_path_info(environ)
+            path = environ['PATH_INFO'].split('/')
+            path.pop(0)  # Pop the scheme.
+            self.path = path
+        elif environ['SCRIPT_NAME'] == '':
+            # Remove the application to set the path.
+            self.application = shift_path_info(environ)
+            self.path = environ['PATH_INFO'].split('/')
+        else:
+            self.application = environ['SCRIPT_NAME']
+            self.path = environ['PATH_INFO'].split('/')
+        self.query_string = environ['QUERY_STRING']
+        return self.handle_request()
+
+    def handle_request(self):
+        """Select the method to handle the request and return a response."""
+        if self.application != 'archive':
+            return self.send_response(httplib.NOT_FOUND)
+        elif self.method == 'PUT':
+            return self.do_PUT()
+        elif self.method == 'POST':
+            return self.do_POST()
+        elif self.method == 'GET':
+            return self.do_GET()
+        return self.send_response(httplib.METHOD_NOT_ALLOWED)
+
+    def send_response(self, code, response='', reason=None):
+        """Set the status code and reason, then return the response."""
+        if reason is None:
+            reason = httplib.responses[code]
+        response_status = '%s %s' % (code, reason)
+        self.start_response(response_status, self.headers.items())
+        return [response]
+
+    def do_PUT(self):
+        """Create an archive or message on PUT."""
+        if len(self.path) == 1:
+            # This expected path is /archive/archive_id.
+            try:
+                self.store.put_archive(self.path[0])
+                return self.send_response(httplib.CREATED)
+            except Exception, error:
+                return self.send_response(
+                    httplib.BAD_REQUEST, reason=error.__doc__)
+        elif len(self.path) == 2:
+            # This expected path is /archive/archive_id/message_id.
+            try:
+                put_input = self.environ['wsgi.input']
+                message = put_input.read(int(self.environ['CONTENT_LENGTH']))
+                self.store.put_message(self.path[0], self.path[1], message)
+                return self.send_response(httplib.CREATED)
+            except Exception, error:
+                return self.send_response(
+                    httplib.BAD_REQUEST, reason=error.__doc__)
+
+    def do_POST(self):
+        """Change a message on POST."""
+        if len(self.path) == 2:
+            # This expected path is /archive/archive_id/message_id.
+            try:
+                # This expected path is /archive/archive_id/message_id.
+                response = self.store.hide_message(
+                    self.path[0], self.path[1], self.query_string)
+                response = simplejson.dumps(response)
+                return self.send_response(httplib.OK, response=response)
+            except Exception, error:
+                return self.send_response(
+                    httplib.BAD_REQUEST, reason=error.__doc__)
+
+    def do_GET(self):
+        """Retrieve a list of messages on GET."""
+        try:
+            response = self.store.get_messages(
+                self.path[0], self.query_string)
+            response = simplejson.dumps(response)
+            return self.send_response(httplib.OK, response=response)
+        except Exception, error:
+            return self.send_response(
+                httplib.BAD_REQUEST, reason=error.__doc__)
+
+    def log_message(self, format, *args):
+        """Override log_message to use standard Python logging."""
+        message = "%s - - [%s] %s\n" % (
+            self.address_string(), self.log_date_time_string(), format % args)
+        self.logger.info(message)
+
+
+class ForkedFakeService:
+    """A Grackle service fake, as a ContextManager."""
+
+    def __init__(self, port, message_archives=None, write_logs=False):
+        """Constructor.
+
+        :param port: The tcp port to use.
+        :param message_archives: A dict of lists of dicts representing
+            archives of messages. The outer dict represents the archive,
+            the list represents the list of messages for that archive.
+        :param write_logs: If true, log messages will be written to stdout.
+        """
+        self.pid = None
+        self.port = port
+        if message_archives is None:
+            self.message_archives = {}
+        else:
+            self.message_archives = message_archives
+        self.read_end, self.write_end = os.pipe()
+        self.write_logs = write_logs
+
+    @staticmethod
+    def from_client(client, message_archives=None):
+        """Instantiate a ForkedFakeService from the client.
+
+        :param port: The client to provide service for.
+        :param message_archives: A dict of lists of dicts representing
+            archives of messages. The outer dict represents the archive,
+            the list represents the list of messages for that archive.
+        """
+        return ForkedFakeService(client.port, message_archives)
+
+    def is_ready(self):
+        """Tell the parent process that the server is ready for writes."""
+        os.write(self.write_end, 'asdf')
+
+    def __enter__(self):
+        """Run the service.
+
+        Fork and start a server in the child.  Return when the server is ready
+        for use."""
+        pid = os.fork()
+        if pid == 0:
+            self.start_server()
+        self.pid = pid
+        os.read(self.read_end, 1)
+        return
+
+    def start_server(self):
+        """Start the HTTP server."""
+        app = GrackleService(MemoryStore(self.message_archives))
+        service = make_server('', self.port, app)
+        self.is_ready()
+        if self.write_logs:
+            logging.basicConfig(
+                stream=sys.stderr, level=logging.INFO)
+        service.serve_forever()
+
+    def __exit__(self, exc_type, exc_val, traceback):
+        os.kill(self.pid, SIGKILL)
+
+
+if __name__ == '__main__':
+    app = GrackleService(MemoryStore({}))
+    service = make_server('', 8787, app)
+    service.serve_forever()

=== added file 'grackle/store.py'
--- grackle/store.py	1970-01-01 00:00:00 +0000
+++ grackle/store.py	2012-03-17 21:36:11 +0000
@@ -0,0 +1,222 @@
+__metaclass__ = type
+__all__ = [
+    'make_json_message',
+    'MemoryStore',
+    'SUPPORTED_DISPLAY_TYPES',
+    'SUPPORTED_ORDERS',
+    ]
+
+import email
+import simplejson
+from urlparse import parse_qs
+
+from grackle.error import (
+    ArchiveIdExists,
+    ArchiveIdNotFound,
+    MessageIdNotFound,
+    UnparsableDateRange,
+    UnsupportedDisplayType,
+    UnsupportedOrder,
+    )
+
+
+SUPPORTED_DISPLAY_TYPES = set(['all', 'text-only', 'headers-only'])
+
+
+SUPPORTED_ORDERS = set(
+    ['date', 'author', 'subject', 'thread_newest', 'thread_oldest',
+     'thread_subject'])
+
+
+def threaded_messages(messages):
+    threads = {}
+    count = 0
+    pending = []
+    for message in messages:
+        if message.get('replies') is None:
+            threads[message['message_id']] = [message]
+            count += 1
+        else:
+            pending.append(message)
+    for message in pending:
+        threads[message['replies']].append(message)
+    return threads.values()
+
+
+def get_body_text(message):
+    """Return the first plain/text messaage part."""
+    if not message.is_multipart():
+        return message.get_payload()
+    for part in email.iterators.typed_subpart_iterator(message, 'multipart'):
+        subparts = part.get_payload()
+        for subpart in subparts:
+            if subpart.get_content_type() == 'text/plain':
+                return subpart.get_payload().strip()
+    return ''
+
+
+def make_json_message(message_id, raw_message, hidden=False):
+    message = email.message_from_string(raw_message)
+    headers = dict(message.items())
+    message = {
+        'message_id': message_id,
+        'headers': headers,
+        # This is broken because the in-reply-to must be encoded.
+        # X-Message-ID-Hash is calculated from the Base 32.
+        'thread_id': headers.get('in-reply-to', message_id),
+        'date': headers.get('date'),
+        'subject': headers.get('subject'),
+        'author': headers.get('from'),
+        'hidden': hidden,
+        'attachments': [],
+        'replies': headers.get('in-reply-to'),
+        'body': get_body_text(message),
+        }
+    return message
+
+
+class MemoryStore:
+    """A memory-backed message store."""
+
+    def __init__(self, message_archives):
+        """Constructor."""
+        self.message_archives = message_archives
+
+    @staticmethod
+    def is_multipart(message):
+        return isinstance(message['body'], list)
+
+    def put_archive(self, archive_id, raw_archive=None):
+        # XXX sinzui 2012-02-29: this needs to raise an error
+        # if the th archive_id is invalid, or the raw archive is not mbox.
+        if archive_id in self.message_archives:
+            raise ArchiveIdExists()
+        self.message_archives[archive_id] = []
+
+    def put_message(self, archive_id, message_id, raw_message):
+        # XXX sinzui 2012-02-29: this needs to raise an error
+        # if the th archive_id is invalid, message_id is not base32
+        # or the raw message is not an email.
+        if archive_id not in self.message_archives:
+            raise ArchiveIdNotFound()
+        if not raw_message:
+            raise ValueError('raw_message is not a message.')
+        json_message = make_json_message(message_id, raw_message)
+        messages = self.message_archives[archive_id]
+        messages.append(json_message)
+
+    def get_messages(self, archive_id, query_string):
+        """Return matching messages.
+
+        :param archive_id: The archive to retrieve from.
+        :param query_string: Contains 'parameters', which is a JSON-format
+            string describing parameters.
+        """
+        query = parse_qs(query_string)
+        parameters = simplejson.loads(query['parameters'][0])
+        order = parameters.get('order')
+        messages = self.message_archives[archive_id]
+        if order is not None:
+            if order not in SUPPORTED_ORDERS:
+                raise UnsupportedOrder
+            elif order.startswith('thread_'):
+                threaded = threaded_messages(messages)
+                messages = []
+                if order == 'thread_subject':
+                    threaded.sort(key=lambda t: t[0]['subject'])
+                if order == 'thread_oldest':
+                    threaded.sort(key=lambda t: min(m['date'] for m in t))
+                if order == 'thread_newest':
+                    threaded.sort(key=lambda t: max(m['date'] for m in t))
+                for thread in threaded:
+                    messages.extend(thread)
+            else:
+                messages.sort(key=lambda m: m[order])
+        display_type = parameters.get('display_type', 'all')
+        if display_type not in SUPPORTED_DISPLAY_TYPES:
+            raise UnsupportedDisplayType
+        if 'date_range' in parameters:
+            try:
+                start_date, end_date = parameters['date_range'].split('..')
+                if not start_date or not end_date:
+                    raise UnparsableDateRange
+            except ValueError:
+                raise UnparsableDateRange
+        new_messages = []
+        for message in messages:
+            if (not parameters['include_hidden'] and message['hidden']):
+                continue
+            if ('message_ids' in parameters
+                and message['message_id'] not in parameters['message_ids']):
+                continue
+            if ('date_range' in parameters
+                and (message['date'] < start_date
+                     or message['date'] > end_date)):
+                continue
+            message = dict(message)
+            if 'headers' in parameters:
+                headers = dict(
+                    (k, v) for k, v in message['headers'].iteritems()
+                    if k in parameters['headers'])
+                message['headers'] = headers
+            if display_type == 'headers-only':
+                del message['body']
+            elif display_type == 'text-only' and self.is_multipart(message):
+                text_parts = [
+                    part.get_payload() for part in message['body']
+                    if part.get_content_type() == 'text/plain']
+                message['body'] = '\n\n'.join(text_parts)
+            elif display_type == 'all' and self.is_multipart(message):
+                parts = [str(part.get_payload()) for part in message['body']]
+                message['body'] = '\n\n'.join(parts)
+            max_body = parameters.get('max_body_length')
+            if max_body is not None and display_type != 'headers-only':
+                message['body'] = message['body'][:max_body]
+            new_messages.append(message)
+        messages = new_messages
+        limit = parameters.get('limit', 100)
+        memo = parameters.get('memo')
+        message_id_indices = dict(
+            (m['message_id'], idx) for idx, m in enumerate(messages))
+        if memo is None:
+            start = 0
+        else:
+            start = message_id_indices[memo.encode('rot13')]
+        if start > 0:
+            previous_memo = messages[start - 1]['message_id'].encode('rot13')
+        else:
+            previous_memo = None
+        end = min(start + limit, len(messages))
+        if end < len(messages):
+            next_memo = messages[end]['message_id'].encode('rot13')
+        else:
+            next_memo = None
+        messages = messages[start:end]
+
+        response = {
+            'messages': messages,
+            'next_memo': next_memo,
+            'previous_memo': previous_memo
+            }
+        return response
+
+    def hide_message(self, archive_id, message_id, query_string):
+        """Change the visbility of a message in an archive.
+
+        :param archive_id: The archive to retrieve from.
+        :param query_string: Contains 'parameters', which is a JSON-format
+            string describing parameters.
+        """
+        query = parse_qs(query_string)
+        parameters = simplejson.loads(query['parameters'][0])
+        hidden = parameters['hidden']
+        messages = self.message_archives[archive_id]
+        for message in messages:
+            if message['message_id'] == message_id:
+                message['hidden'] = hidden
+            response = {
+                'message_id': message_id,
+                'hidden': hidden,
+                }
+            return response
+        raise MessageIdNotFound

=== added directory 'grackle/testing'
=== added file 'grackle/testing/__init__.py'
=== added file 'grackle/testing/cassandra.py'
--- grackle/testing/cassandra.py	1970-01-01 00:00:00 +0000
+++ grackle/testing/cassandra.py	2012-01-22 10:39:15 +0000
@@ -0,0 +1,54 @@
+# Copyright (c) 2011 Canonical Ltd
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Affero General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Affero General Public License for more details.
+#
+# You should have received a copy of the GNU Affero General Public
+# License along with this program. If not, see
+# <http://www.gnu.org/licenses/>.
+
+"""Test helpers for working with cassandra."""
+
+import os.path
+
+from fixtures import Fixture, TempDir
+from pycassa.system_manager import SystemManager
+
+from grackle.cassandra import workaround_1779
+from grackle.model import create_schema
+
+
+class TemporaryKeyspace(Fixture):
+    """Create a temporary keyspace.
+
+    The keyspace is named after a tempdir.
+    """
+
+    def setUp(self):
+        super(TemporaryKeyspace, self).setUp()
+        tempdir = self.useFixture(TempDir())
+        self.keyspace = os.path.basename(tempdir.path)
+        self.mgr = SystemManager()
+        workaround_1779(self.mgr.create_keyspace, self.keyspace,
+            replication_factor=1)
+        self.addCleanup(workaround_1779, self.mgr.drop_keyspace,
+            self.keyspace)
+
+
+class TemporaryDB(Fixture):
+    """Create a temporary usable OOPS DB.
+
+    The keyspace for it is at self.keyspace.
+    """
+
+    def setUp(self):
+        super(TemporaryDB, self).setUp()
+        self.keyspace = self.useFixture(TemporaryKeyspace()).keyspace
+        create_schema('localhost', self.keyspace)

=== modified file 'grackle/tests/__init__.py'
--- grackle/tests/__init__.py	2012-01-10 11:16:56 +0000
+++ grackle/tests/__init__.py	2012-01-22 10:39:15 +0000
@@ -0,0 +1,28 @@
+# Copyright (c) 2012 Canonical Ltd
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Affero General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Affero General Public License for more details.
+#
+# You should have received a copy of the GNU Affero General Public
+# License along with this program. If not, see
+# <http://www.gnu.org/licenses/>.
+
+from unittest import TestLoader
+
+
+def test_suite():
+    test_names = [
+        'model',
+        'client',
+        'wsgi',
+        ]
+    tests = ['grackle.tests.test_' + test for test in test_names]
+    loader = TestLoader()
+    return loader.loadTestsFromNames(tests)

=== modified file 'grackle/tests/test_client.py'
--- grackle/tests/test_client.py	2012-01-30 18:28:17 +0000
+++ grackle/tests/test_client.py	2012-03-17 21:36:11 +0000
@@ -1,242 +1,91 @@
-from BaseHTTPServer import (
-    HTTPServer,
-    BaseHTTPRequestHandler,
-    )
-import httplib
-import logging
-import os
-from signal import SIGKILL
-import simplejson
+__metaclass__ = type
+
+from email.message import Message
+from email.mime.multipart import MIMEMultipart
+from email.mime.text import MIMEText
 from StringIO import StringIO
-import sys
 from unittest import TestCase
-from urlparse import urlparse
-from urlparse import parse_qs
 
 from testtools import ExpectedException
 
-from grackle.client import (
-    GrackleClient,
+from grackle.client import GrackleClient
+from grackle.error import (
+    ArchiveIdExists,
+    UnparsableDateRange,
+    UnsupportedDisplayType,
     UnsupportedOrder,
     )
-
-
-def threaded_messages(messages):
-    threads = {}
-    count = 0
-    pending = []
-    for message in messages:
-        if message.get('in_reply_to') is None:
-            threads[message['message_id']] = [message]
-            count += 1
-        else:
-            pending.append(message)
-    for message in pending:
-        threads[message['in_reply_to']].append(message)
-    return threads.values()
-
-
-class GrackleStore:
-    """A memory-backed message store."""
-
-    def __init__(self, messages):
-        """Constructor."""
-        self.messages = messages
-
-    def get_messages(self, archive_id, query_string):
-        """Return matching messages.
-
-        :param archive_id: The archive to retrieve from.
-        :param query_string: Contains 'parameters', which is a JSON-format
-            string describing parameters.
-        """
-        query = parse_qs(query_string)
-        parameters = simplejson.loads(query['parameters'][0])
-        order = parameters.get('order')
-        messages = self.messages[archive_id]
-        if order is not None:
-            if order not in SUPPORTED_ORDERS:
-                raise UnsupportedOrder
-            elif order.startswith('thread_'):
-                threaded = threaded_messages(messages)
-                messages = []
-                if order == 'thread_subject':
-                    threaded.sort(key=lambda t: t[0]['subject'])
-                if order == 'thread_oldest':
-                    threaded.sort(key=lambda t: min(m['date'] for m in t))
-                if order == 'thread_newest':
-                    threaded.sort(key=lambda t: max(m['date'] for m in t))
-                for thread in threaded:
-                    messages.extend(thread)
-            else:
-                messages.sort(key=lambda m: m[order])
-        new_messages = []
-        for message in messages:
-            if (not parameters['include_hidden']
-                and message.get('hidden', False)):
-                continue
-
-            if ('message_ids' in parameters
-                and message['message_id'] not in parameters['message_ids']):
-                continue
-            message = dict(message)
-            if 'headers' in parameters:
-                headers = dict(
-                    (k, v) for k, v in message['headers'].iteritems()
-                    if k in parameters['headers'])
-                message['headers'] = headers
-            max_body = parameters.get('max_body_length')
-            if max_body is not None:
-                message['body'] = message['body'][:max_body]
-            new_messages.append(message)
-        messages = new_messages
-        limit = parameters.get('limit', 100)
-        memo = parameters.get('memo')
-        message_id_indices = dict(
-            (m['message_id'], idx) for idx, m in enumerate(messages))
-        if memo is None:
-            start = 0
-        else:
-            start = message_id_indices[memo.encode('rot13')]
-        if start > 0:
-            previous_memo = messages[start - 1]['message_id'].encode('rot13')
-        else:
-            previous_memo = None
-        end = min(start + limit, len(messages))
-        if end < len(messages):
-            next_memo = messages[end]['message_id'].encode('rot13')
-        else:
-            next_memo = None
-        messages = messages[start:end]
-
-        response = {
-            'messages': messages,
-            'next_memo': next_memo,
-            'previous_memo': previous_memo
-            }
-        return response
-
-
-class ForkedFake:
-    """A Grackle service fake, as a ContextManager."""
-
-    def __init__(self, port, messages=None, write_logs=False):
-        """Constructor.
-        :param port: The tcp port to use
-        :param messages: A dict of lists of dicts representing messages.  The
-            outer dict represents the archive, the list represents the list of
-            messages for that archive.
-        :param write_logs: If true, log messages will be written to stdout.
-        """
-        self.pid = None
-        self.port = port
-        if messages is None:
-            self.messages = {}
-        else:
-            self.messages = messages
-        self.read_end, self.write_end = os.pipe()
-        self.write_logs = write_logs
-
-    @staticmethod
-    def from_client(client, messages=None):
-        """Instantiate a ForkedFake from the client.
-
-        :param port: The client  to provide service for.
-        :param messages: A dict of lists of dicts representing messages.  The
-            outer dict represents the archive, the list represents the list of
-            messages for that archive.
-        """
-        return ForkedFake(client.port, messages)
-
-    def is_ready(self):
-        """Tell the parent process that the server is ready for writes."""
-        os.write(self.write_end, 'asdf')
-
-    def __enter__(self):
-        """Run the service.
-
-        Fork and start a server in the child.  Return when the server is ready
-        for use."""
-        pid = os.fork()
-        if pid == 0:
-            self.start_server()
-        self.pid = pid
-        os.read(self.read_end, 1)
-        return
-
-    def start_server(self):
-        """Start the HTTP server."""
-        service = HTTPServer(('', self.port), FakeGrackleRequestHandler)
-        service.store = GrackleStore(self.messages)
-        for archive_id, messages in service.store.messages.iteritems():
-            for message in messages:
-                message.setdefault('headers', {})
-        self.is_ready()
-        if self.write_logs:
-            logging.basicConfig(
-                stream=sys.stderr, level=logging.INFO)
-        service.serve_forever()
-
-    def __exit__(self, exc_type, exc_val, traceback):
-        os.kill(self.pid, SIGKILL)
-
-
-SUPPORTED_ORDERS = set(
-    ['date', 'author', 'subject', 'thread_newest', 'thread_oldest',
-     'thread_subject'])
-
-
-class FakeGrackleRequestHandler(BaseHTTPRequestHandler):
-    """A request handler that forwards to server.store."""
-
-    def __init__(self, *args, **kwargs):
-        """Constructor.  Sets up logging."""
-        self.logger = logging.getLogger('http')
-        BaseHTTPRequestHandler.__init__(self, *args, **kwargs)
-
-    def do_POST(self):
-        """Create a message on POST."""
-        message = self.rfile.read(int(self.headers['content-length']))
-        if message == 'This is a message':
-            self.send_response(httplib.CREATED)
-            self.end_headers()
-            self.wfile.close()
-        else:
-            self.send_error(httplib.BAD_REQUEST)
-
-    def do_GET(self):
-        """Retrieve a list of messages on GET."""
-        scheme, netloc, path, params, query_string, fragments = (
-            urlparse(self.path))
-        parts = path.split('/')
-        if parts[1] == 'archive':
-            try:
-                response = self.server.store.get_messages(
-                    parts[2], query_string)
-                self.send_response(httplib.OK)
-                self.end_headers()
-                self.wfile.write(simplejson.dumps(response))
-            except UnsupportedOrder:
-                self.send_response(httplib.BAD_REQUEST)
-                self.wfile.write('Unsupported order')
-                return
-
-    def log_message(self, format, *args):
-        """Override log_message to use standard Python logging."""
-        message = "%s - - [%s] %s\n" % (
-            self.address_string(), self.log_date_time_string(), format % args)
-        self.logger.info(message)
+from grackle.service import ForkedFakeService
+from grackle.store import make_json_message
+
+
+def make_message(message_id, body='body', headers=None, hidden=False):
+    if headers is None:
+        headers = {}
+    message_headers = {
+        'Message-Id': message_id,
+        'date': '2005-01-01',
+        'subject': 'subject',
+        'from': 'author',
+        'replies': '',
+        }
+    message_headers.update(headers.items())
+    message = Message()
+    message.set_payload(body)
+    for key, value in message_headers.items():
+        message[key] = value
+    return make_json_message(message_id, message.as_string(), hidden)
+
+
+def make_mime_message(message_id, body='body', headers=None, hidden=False,
+                      attachment_type=None):
+    parts = MIMEMultipart()
+    parts.attach(MIMEText(body))
+    if attachment_type is not None:
+        attachment = Message()
+        attachment.set_payload('attactment data.')
+        attachment['Content-Type'] = attachment_type
+        attachment['Content-Disposition'] = 'attachment; filename="file.ext"'
+        parts.attach(attachment)
+    return make_message(message_id, parts.as_string(), headers, hidden)
+
+
+class TestPutArchive(TestCase):
+
+    def test_put_archive(self):
+        client = GrackleClient('localhost', 8410)
+        message_archives = {}
+        with ForkedFakeService.from_client(client, message_archives):
+            client.put_archive('arch1')
+            response = client.get_messages('arch1')
+        self.assertEqual(0, len(response['messages']))
+
+    def test_put_archive_existing_archive(self):
+        client = GrackleClient('localhost', 8411)
+        message_archives = {'arch1': []}
+        with ForkedFakeService.from_client(client, message_archives):
+            with ExpectedException(ArchiveIdExists, ''):
+                client.put_archive('arch1')
 
 
 class TestPutMessage(TestCase):
 
     def test_put_message(self):
-        client = GrackleClient('localhost', 8436)
-        with ForkedFake.from_client(client):
-            client.put_message('arch1', 'asdf', StringIO('This is a message'))
+        client = GrackleClient('localhost', 8420)
+        message_archives = {'arch1': []}
+        with ForkedFakeService.from_client(client, message_archives):
+            client.put_message('arch1', 'id1', StringIO('This is a message'))
+            response = client.get_messages('arch1')
+        self.assertEqual(1, len(response['messages']))
+        message = response['messages'][0]
+        self.assertEqual('id1', message['message_id'])
+
+    def test_put_message_without_archive(self):
+        client = GrackleClient('localhost', 8421)
+        message_archives = {'arch1': []}
+        with ForkedFakeService.from_client(client, message_archives):
             with ExpectedException(Exception, 'wtf'):
-                client.put_message('arch1', 'asdf',
-                    StringIO('This is not a message'))
+                client.put_message('no-archive', 'id1', StringIO('message'))
 
 
 class TestGetMessages(TestCase):
@@ -249,11 +98,10 @@
             sorted(ids), sorted(messages, key=lambda m: m['message_id']))
 
     def test_get_messages(self):
-        client = GrackleClient('localhost', 8435)
-        with ForkedFake.from_client(client,
-            {'baz':
-            [{'message_id': 'foo'},
-             {'message_id': 'bar'}]}):
+        client = GrackleClient('localhost', 8430)
+        archive = {
+            'baz': [make_message('foo'), make_message('bar')]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz')
         self.assertEqual(['bar', 'foo'], sorted(m['message_id'] for m in
             response['messages']))
@@ -262,20 +110,17 @@
 
     def test_get_messages_by_id(self):
         client = GrackleClient('localhost', 8437)
-        with ForkedFake.from_client(client,
-            {'baz':
-            [{'message_id': 'foo'},
-             {'message_id': 'bar'}]}):
+        archive = {
+            'baz': [make_message('foo'), make_message('bar')]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz', message_ids=['foo'])
         message, = response['messages']
         self.assertEqual('foo', message['message_id'])
 
     def test_get_messages_batching(self):
         client = GrackleClient('localhost', 8438)
-        with ForkedFake.from_client(client,
-            {'baz':
-            [{'message_id': 'foo'},
-             {'message_id': 'bar'}]}):
+        archive = {'baz': [make_message('foo'), make_message('bar')]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz', limit=1)
             self.assertEqual(1, len(response['messages']))
             messages = response['messages']
@@ -287,9 +132,16 @@
 
     def get_messages_member_order_test(self, key):
         client = GrackleClient('localhost', 8439)
-        with ForkedFake.from_client(client,
-                {'baz': [{'message_id': 'foo', key: '2011-03-25'},
-                 {'message_id': 'bar', key: '2011-03-24'}]}):
+        if key == 'author':
+            header_name = 'from'
+        else:
+            header_name = key
+        archive = {
+            'baz': [
+                make_message('foo', headers={header_name: '2011-03-25'}),
+                make_message('bar', headers={header_name: '2011-03-24'}),
+             ]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz')
             self.assertIDOrder(['foo', 'bar'], response['messages'])
             response = client.get_messages('baz', order=key)
@@ -305,12 +157,15 @@
         self.get_messages_member_order_test('subject')
 
     def test_get_messages_thread_subject_order(self):
+        archive = {
+            'baz': [
+                make_message('bar', headers={'subject': 'y'}),
+                make_message('qux', headers={'subject': 'z'}),
+                make_message('foo', headers={'subject': 'x',
+                                             'in-reply-to': 'qux'}),
+             ]}
         client = GrackleClient('localhost', 8439)
-        with ForkedFake.from_client(client, {'baz': [
-            {'message_id': 'bar', 'subject': 'y'},
-            {'message_id': 'qux', 'subject': 'z'},
-            {'message_id': 'foo', 'subject': 'x', 'in_reply_to': 'qux'},
-            ]}):
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz')
             self.assertIDOrder(['bar', 'qux', 'foo'], response['messages'])
             response = client.get_messages('baz', order='subject')
@@ -320,11 +175,14 @@
 
     def test_get_messages_thread_oldest_order(self):
         client = GrackleClient('localhost', 8439)
-        with ForkedFake.from_client(client, {'baz': [
-            {'message_id': 'bar', 'date': 'x'},
-            {'message_id': 'qux', 'date': 'z'},
-            {'message_id': 'foo', 'date': 'y', 'in_reply_to': 'qux'},
-            ]}):
+        archive = {
+            'baz': [
+                make_message('bar', headers={'date': 'x'}),
+                make_message('qux', headers={'date': 'z'}),
+                make_message('foo', headers={'date': 'y',
+                                             'in-reply-to': 'qux'}),
+            ]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz')
             self.assertIDOrder(['bar', 'qux', 'foo'], response['messages'])
             response = client.get_messages('baz', order='date')
@@ -334,12 +192,16 @@
 
     def test_get_messages_thread_newest_order(self):
         client = GrackleClient('localhost', 8439)
-        with ForkedFake.from_client(client, {'baz': [
-            {'message_id': 'bar', 'date': 'x'},
-            {'message_id': 'qux', 'date': 'w'},
-            {'message_id': 'foo', 'date': 'y', 'in_reply_to': 'bar'},
-            {'message_id': 'baz', 'date': 'z', 'in_reply_to': 'qux'},
-            ]}):
+        archive = {
+            'baz': [
+                make_message('bar', headers={'date': 'x'}),
+                make_message('qux', headers={'date': 'w'}),
+                make_message('foo', headers={'date': 'y',
+                                             'in-reply-to': 'bar'}),
+                make_message('baz', headers={'date': 'z',
+                                             'in-reply-to': 'qux'}),
+            ]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz', order='date')
             self.assertIDOrder(
                 ['qux', 'bar', 'foo', 'baz'], response['messages'])
@@ -349,18 +211,19 @@
 
     def test_get_messages_unsupported_order(self):
         client = GrackleClient('localhost', 8439)
-        with ForkedFake.from_client(client,
-                {'baz': [{'message_id': 'foo', 'date': '2011-03-25'},
-                 {'message_id': 'bar', 'date': '2011-03-24'}]}):
+        archive = {
+            'baz': [
+                make_message('foo', headers={'date': '2011-03-25'}),
+                make_message('foo', headers={'date': '2011-03-24'}),
+            ]}
+        with ForkedFakeService.from_client(client, archive):
             with ExpectedException(UnsupportedOrder, ''):
                 client.get_messages('baz', order='nonsense')
 
     def test_get_messages_headers_no_headers(self):
         client = GrackleClient('localhost', 8440)
-        with ForkedFake.from_client(client,
-            {'baz': [
-                {'message_id': 'foo'}
-            ]}):
+        archive = {'baz': [make_message('foo')]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz', headers=[
                 'Subject', 'Date', 'X-Launchpad-Message-Rationale'])
         first_message = response['messages'][0]
@@ -369,10 +232,9 @@
 
     def test_get_messages_headers_exclude_headers(self):
         client = GrackleClient('localhost', 8441)
-        with ForkedFake.from_client(client,
-            {'baz': [
-                {'message_id': 'foo', 'headers': {'From': 'me'}}
-            ]}):
+        archive = {
+            'baz': [make_message('foo', headers={'From': 'me'})]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz', headers=[
                 'Subject', 'Date', 'X-Launchpad-Message-Rationale'])
         first_message = response['messages'][0]
@@ -381,10 +243,10 @@
 
     def test_get_messages_headers_include_headers(self):
         client = GrackleClient('localhost', 8442)
-        with ForkedFake.from_client(client,
-            {'baz': [
-                {'message_id': 'foo', 'headers': {'From': 'me', 'To': 'you'}}
-            ]}):
+        archive = {
+            'baz': [
+                make_message('foo', headers={'From': 'me', 'To': 'you'})]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz', headers=[
                 'From', 'To'])
         first_message = response['messages'][0]
@@ -393,22 +255,141 @@
 
     def test_get_messages_max_body_length(self):
         client = GrackleClient('localhost', 8443)
-        with ForkedFake.from_client(client,
-            {'baz': [
-                {'message_id': 'foo', 'body': u'abcdefghi'}
-            ]}):
+        archive = {'baz': [make_message('foo', body=u'abcdefghi')]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz', max_body_length=3)
         first_message = response['messages'][0]
         self.assertEqual('abc', first_message['body'])
 
     def test_include_hidden(self):
         client = GrackleClient('localhost', 8444)
-        with ForkedFake.from_client(client,
-            {'baz': [
-                {'message_id': 'foo', 'hidden': True},
-                {'message_id': 'bar', 'hidden': False}
-            ]}):
+        archive = {
+            'baz': [
+                make_message('foo', hidden=True),
+                make_message('bar', hidden=False),
+            ]}
+        with ForkedFakeService.from_client(client, archive):
             response = client.get_messages('baz', include_hidden=True)
             self.assertMessageIDs(['bar', 'foo'], response['messages'])
             response = client.get_messages('baz', include_hidden=False)
             self.assertMessageIDs(['bar'], response['messages'])
+
+    def test_display_type_unknown_value(self):
+        client = GrackleClient('localhost', 8445)
+        archive = {'baz': [make_message('foo', body=u'abcdefghi')]}
+        with ForkedFakeService.from_client(client, archive):
+            with ExpectedException(UnsupportedDisplayType, ''):
+                client.get_messages('baz', display_type='unknown')
+
+    def test_display_type_headers_only(self):
+        client = GrackleClient('localhost', 8446)
+        archive = {
+            'baz': [
+                make_message('foo', body=u'abcdefghi',
+                             headers={'From': 'me', 'To': 'you'})]}
+        with ForkedFakeService.from_client(client, archive):
+            response = client.get_messages('baz', display_type='headers-only')
+        first_message = response['messages'][0]
+        self.assertEqual('foo', first_message['message_id'])
+        self.assertEqual(
+            archive['baz'][0]['headers'], first_message['headers'])
+        self.assertNotIn('body', first_message)
+
+    def test_display_type_text_only(self):
+        client = GrackleClient('localhost', 8446)
+        archive = {
+            'baz': [
+                make_mime_message(
+                    'foo', 'abcdefghi',
+                    headers={'From': 'me', 'To': 'you'},
+                    attachment_type='text/x-diff')]}
+        with ForkedFakeService.from_client(client, archive):
+            response = client.get_messages('baz', display_type='text-only')
+        first_message = response['messages'][0]
+        self.assertEqual('foo', first_message['message_id'])
+        self.assertEqual('me', first_message['headers']['From'])
+        self.assertEqual('you', first_message['headers']['To'])
+        self.assertEqual(archive['baz'][0]['body'], first_message['body'])
+
+    def test_display_type_all(self):
+        client = GrackleClient('localhost', 8447)
+        archive = {
+            'baz': [
+                make_mime_message(
+                    'foo', 'abcdefghi',
+                    headers={'From': 'me', 'To': 'you'},
+                    attachment_type='text/x-diff')]}
+        with ForkedFakeService.from_client(client, archive):
+            response = client.get_messages('baz', display_type='all')
+        first_message = response['messages'][0]
+        self.assertEqual('foo', first_message['message_id'])
+        self.assertEqual('me', first_message['headers']['From'])
+        self.assertEqual('you', first_message['headers']['To'])
+        self.assertEqual(archive['baz'][0]['body'], first_message['body'])
+
+    def test_date_range(self):
+        client = GrackleClient('localhost', 8448)
+        archive = {
+            'baz': [
+                make_mime_message(
+                    'foo', 'abcdefghi', headers={'date': '2011-12-31'}),
+                make_mime_message(
+                    'bar', 'abcdefghi', headers={'date': '2012-01-01'}),
+                make_mime_message(
+                    'qux', 'abcdefghi', headers={'date': '2012-01-15'}),
+                make_mime_message(
+                    'naf', 'abcdefghi', headers={'date': '2012-01-31'}),
+                make_mime_message(
+                    'doh', 'abcdefghi', headers={'date': '2012-02-01'}),
+                    ]}
+        with ForkedFakeService.from_client(client, archive):
+            response = client.get_messages(
+                'baz', date_range='2012-01-01..2012-01-31')
+        ids = sorted(m['message_id'] for m in response['messages'])
+        self.assertEqual(['bar', 'naf', 'qux'], ids)
+
+    def test_date_range_unparsabledaterange(self):
+        client = GrackleClient('localhost', 8449)
+        archive = {'baz': [make_message('foo', body=u'abcdefghi')]}
+        with ForkedFakeService.from_client(client, archive):
+            with ExpectedException(UnparsableDateRange, ''):
+                client.get_messages('baz', date_range='2012-01-01')
+
+    def test_date_range_unparsabledaterange_missing_part(self):
+        client = GrackleClient('localhost', 8450)
+        archive = {'baz': [make_message('foo', body=u'abcdefghi')]}
+        with ForkedFakeService.from_client(client, archive):
+            with ExpectedException(UnparsableDateRange, ''):
+                client.get_messages('baz', date_range='2012-01-01..')
+
+    def test_date_range_unparsabledaterange_extra_part(self):
+        client = GrackleClient('localhost', 8451)
+        archive = {'baz': [make_message('foo', body=u'abcdefghi')]}
+        with ForkedFakeService.from_client(client, archive):
+            with ExpectedException(UnparsableDateRange, ''):
+                client.get_messages('baz', date_range='2012-01..12-02..12-03')
+
+
+class TestHideMessages(TestCase):
+
+    def test_hide_message_true(self):
+        client = GrackleClient('localhost', 8470)
+        archive = {
+            'baz': [
+                make_message('foo', hidden=False),
+            ]}
+        with ForkedFakeService.from_client(client, archive):
+            response = client.hide_message('baz', 'foo', hidden=True)
+        self.assertEqual('foo', response['message_id'])
+        self.assertIs(True, response['hidden'])
+
+    def test_hide_message_false(self):
+        client = GrackleClient('localhost', 8470)
+        archive = {
+            'baz': [
+                make_message('foo', hidden=True),
+            ]}
+        with ForkedFakeService.from_client(client, archive):
+            response = client.hide_message('baz', 'foo', hidden=False)
+        self.assertEqual('foo', response['message_id'])
+        self.assertIs(False, response['hidden'])

=== added file 'grackle/tests/test_model.py'
--- grackle/tests/test_model.py	1970-01-01 00:00:00 +0000
+++ grackle/tests/test_model.py	2012-01-24 22:54:07 +0000
@@ -0,0 +1,225 @@
+# Copyright (c) 2012 Canonical Ltd
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Affero General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Affero General Public License for more details.
+#
+# You should have received a copy of the GNU Affero General Public
+# License along with this program. If not, see
+# <http://www.gnu.org/licenses/>.
+
+__metaclass__ = type
+
+import calendar
+import datetime
+from email.utils import formatdate
+import os
+from tempfile import _RandomNameSequence
+
+from dateutil.tz import (
+    tzoffset,
+    tzutc,
+    )
+from pycassa.util import convert_uuid_to_time
+import testtools
+
+from grackle.model import (
+    CassandraConnection,
+    _format_message,
+    _parse_message,
+    )
+
+
+TEMPLATE_MESSAGE = """\
+From: sysadmin@example.com
+To: developer@example.com
+Subject: Everything is broken
+Date: {date}
+Message-Id: {id}
+
+Help, everything has just broken.
+"""
+
+TEST_MESSAGE = TEMPLATE_MESSAGE.format(
+    date='Sat, 1 Jan 2000 11:02:34 +1100',
+    id='<aaaaaaaaaaaaa@example.com>')
+
+
+class TestParseMessage(testtools.TestCase):
+
+    def test_works(self):
+        # _parse_message extracts interesting fields. It also parses the
+        # date and returns it separately.
+        date, msg = _parse_message(TEST_MESSAGE)
+        self.assertEqual('sysadmin@example.com', msg['from'])
+        self.assertEqual('developer@example.com', msg['to'])
+        self.assertEqual('Everything is broken', msg['subject'])
+        self.assertEqual('2000-01-01T11:02:34+11:00', msg['date'])
+        self.assertEqual('<aaaaaaaaaaaaa@example.com>', msg['message-id'])
+        self.assertEqual(
+            datetime.datetime(
+                2000, 1, 1, 11, 2, 34, tzinfo=tzoffset('', 39600)),
+            date)
+
+
+class TestAddMessage(testtools.TestCase):
+
+    def test_add_message(self):
+        c = CassandraConnection(
+            os.environ['GRACKLE_TEST_KEYSPACE'], ['localhost:9160'])
+        archive = next(_RandomNameSequence())
+
+        # Write the message out to Cassandra, and read it back in.
+        key = c.add_message(archive, TEST_MESSAGE)
+        cmsg = c.messages.get(key)
+
+        # The archive should contain a single message, a reference to
+        # our new key.
+        archive_messages = c.archive_messages.get(archive).items()
+        self.assertEqual(1, len(archive_messages))
+        self.assertEqual(key, archive_messages[0][1])
+
+        # The key in archive_message is a TimeUUID for the Date field in
+        # the message.
+        utctime = calendar.timegm(datetime.datetime(
+            2000, 1, 1, 0, 2, 34, tzinfo=tzutc()).timetuple())
+        self.assertEqual(
+            utctime,
+            convert_uuid_to_time(archive_messages[0][0]))
+
+        # The stored message contains the full original text of the
+        # message, as well as interesting fields parsed out.
+        self.assertEqual(TEST_MESSAGE, cmsg['raw'])
+        parsed_message = _parse_message(TEST_MESSAGE)[1]
+        for key, value in parsed_message.iteritems():
+            self.assertEqual(value, cmsg[key])
+
+
+class TestGetMessages(testtools.TestCase):
+
+    def assertMessages(self, expected_ids, messages):
+        expected_msgids = [
+            '<message%d@example.com>' % id for id in expected_ids]
+        actual_msgids = [msg['headers']['message-id'] for msg in messages]
+        self.assertEqual(expected_msgids, actual_msgids)
+
+    def makeMessages(self, conn, archive, count):
+        return [
+            conn.add_message(
+                archive,
+                TEMPLATE_MESSAGE.format(
+                    date=formatdate(i * 100),
+                    id='<message%d@example.com>' % i))
+            for i in range(count)]
+
+    def makeArchive(self):
+        conn = CassandraConnection(
+            os.environ['GRACKLE_TEST_KEYSPACE'], ['localhost:9160'])
+        archive = next(_RandomNameSequence())
+        return conn, archive
+
+    def test_single_message(self):
+        conn, archive = self.makeArchive()
+        self.makeMessages(conn, archive, 1)
+
+        # We get a single message when we ask for it.
+        messages = conn.get_messages(archive, 'date', 1, '')[1]
+        self.assertMessages([0], messages)
+        cmsg = messages[0]
+
+        # The raw message matches what we generated, and the result is
+        # formatted according to the default settings.
+        expected_raw = TEMPLATE_MESSAGE.format(
+            date=formatdate(0), id='<message0@example.com>')
+        self.assertEqual(expected_raw, messages[0]['raw'])
+        pmsg = _parse_message(expected_raw)[1]
+        pmsg['raw'] = expected_raw
+        self.assertEqual(
+            _format_message(
+                pmsg,
+                headers=['date', 'from', 'subject', 'message-id'],
+                include_raw=True),
+            cmsg)
+
+    def test_limit(self):
+        conn, archive = self.makeArchive()
+        self.makeMessages(conn, archive, 4)
+        self.assertMessages(
+            [0, 1], conn.get_messages(archive, 'date', 2, '')[1])
+
+    def test_order(self):
+        conn, archive = self.makeArchive()
+        self.makeMessages(conn, archive, 4)
+        self.assertMessages(
+            [3, 2], conn.get_messages(archive, '-date', 2, '')[1])
+
+    def test_batching_forward(self):
+        conn, archive = self.makeArchive()
+        self.makeMessages(conn, archive, 5)
+        prev, messages, next = conn.get_messages(archive, 'date', 2, '')
+        self.assertMessages([0, 1], messages)
+        prev, messages, next = conn.get_messages(archive, 'date', 2, next)
+        self.assertMessages([2, 3], messages)
+        prev, messages, next = conn.get_messages(archive, 'date', 2, next)
+        self.assertMessages([4], messages)
+        prev, messages, next = conn.get_messages(archive, 'date', 2, next)
+        self.assertIs(None, prev)
+        self.assertMessages([], messages)
+        self.assertIs(None, next)
+
+    def test_batching_backward(self):
+        conn, archive = self.makeArchive()
+        self.makeMessages(conn, archive, 5)
+        prev, messages, next = conn.get_messages(archive, 'date', 2, '')
+        self.assertMessages([0, 1], messages)
+        prev, messages, next = conn.get_messages(archive, 'date', 2, next)
+        self.assertMessages([2, 3], messages)
+        prev, messages, next = conn.get_messages(
+            archive, 'date', 2, prev, backward=True)
+        self.assertMessages([0, 1], messages)
+        prev, messages, next = conn.get_messages(
+            archive, 'date', 2, prev, backward=True)
+        self.assertIs(None, prev)
+        self.assertMessages([], messages)
+        self.assertIs(None, next)
+
+    def test_date_filter(self):
+        conn, archive = self.makeArchive()
+        self.makeMessages(conn, archive, 10)
+        start = datetime.datetime.utcfromtimestamp(250).replace(
+            tzinfo=tzutc())
+        finish = datetime.datetime.utcfromtimestamp(500).replace(
+            tzinfo=tzutc())
+        prev, messages, next = conn.get_messages(
+            archive, 'date', 2, '', start_date=start, finish_date=finish)
+        self.assertMessages([3, 4], messages)
+        prev, messages, next = conn.get_messages(
+            archive, 'date', 2, next, start_date=start, finish_date=finish)
+        self.assertMessages([5], messages)
+
+
+class TestMessageFormatter(testtools.TestCase):
+
+    def test_all(self):
+        parsed = _parse_message(TEST_MESSAGE)[1]
+        parsed['raw'] = TEST_MESSAGE
+        formatted = _format_message(
+            parsed,
+            headers=['date', 'from', 'subject', 'message-id'],
+            include_raw=True)
+        expected = {
+            'headers': {
+                'date': '2000-01-01T11:02:34+11:00',
+                'from': 'sysadmin@example.com',
+                'message-id': '<aaaaaaaaaaaaa@example.com>',
+                'subject': 'Everything is broken',
+                },
+            'raw': TEST_MESSAGE,
+            }
+        self.assertEqual(expected, formatted)

=== added file 'grackle/tests/test_wsgi.py'
--- grackle/tests/test_wsgi.py	1970-01-01 00:00:00 +0000
+++ grackle/tests/test_wsgi.py	2012-01-22 10:39:15 +0000
@@ -0,0 +1,41 @@
+# Copyright (c) 2012 Canonical Ltd
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Affero General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Affero General Public License for more details.
+#
+# You should have received a copy of the GNU Affero General Public
+# License along with this program. If not, see
+# <http://www.gnu.org/licenses/>.
+
+__metaclass__ = type
+
+from unittest import TestCase
+
+from webtest import TestApp
+
+from grackle.wsgi import app
+
+
+class TestRoot(TestCase):
+
+    def test_get(self):
+        test_app = TestApp(app.wsgifunc())
+        r = test_app.get('/')
+        self.assertEqual(200, r.status_int)
+        r.mustcontain('Service endpoint for Grackle.')
+
+
+class TestArchive(TestCase):
+
+    def test_get(self):
+        test_app = TestApp(app.wsgifunc())
+        r = test_app.get('/archive/foo')
+        self.assertEqual(200, r.status_int)
+        r.mustcontain('Service for archive foo.')

=== added file 'grackle/wsgi.py'
--- grackle/wsgi.py	1970-01-01 00:00:00 +0000
+++ grackle/wsgi.py	2012-01-22 12:36:48 +0000
@@ -0,0 +1,86 @@
+# Copyright (c) 2011 Canonical Ltd
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Affero General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Affero General Public License for more details.
+#
+# You should have received a copy of the GNU Affero General Public
+# License along with this program. If not, see
+# <http://www.gnu.org/licenses/>.
+
+__metaclass__ = type
+
+import json
+import uuid
+
+import web
+
+from grackle.model import CassandraConnection
+
+conn = None
+
+
+class root:
+
+    def GET(self):
+        web.header('Content-Type', 'text/plain')
+        return (
+            'Service endpoint for Grackle.\n\n'
+            'When you want mail that was archived in Grackle, you can get it '
+            'from Grackle')
+
+
+class archive:
+
+    def GET(self, name):
+        web.header('Content-Type', 'text/plain')
+        args = web.input()
+        if not getattr(args, 'op', None):
+            return self.index(name)
+
+        if args.op == 'get_messages':
+            return self.get_messages(name, args)
+
+    def POST(self, name):
+        data = web.data()
+        conn.add_message(name, data)
+        web.ctx.status = '201 Created'
+
+    def index(self, name):
+        return ('Service for archive %s.' % name)
+
+    def get_messages(self, name, args):
+        try:
+            count = int(args.count)
+        except (AttributeError, ValueError):
+            count = 10
+        order = getattr(args, 'order', '-date')
+        memo = getattr(args, 'memo', '')
+        prev, messages, next = conn.get_messages(name, order, count, memo)
+        return json.dumps({
+            'prev_memo': prev,
+            'messages': messages,
+            'next_memo': next,
+            })
+
+
+urls = (
+    '/', 'root',
+    '/archive/(.+)', 'archive',
+    )
+
+app = web.application(urls, globals(), autoreload=False)
+
+if __name__ == "__main__":
+    import logging
+    logging.basicConfig(
+        format='%(asctime)s:%(levelname)s:%(name)s:%(message)s',
+        level=logging.DEBUG)
+    conn = CassandraConnection('grackle', ['localhost:9160'])
+    app.run()

